/*
 * eth_ecat_sdo_mailbox.cpp
 *
 *  Created on: May 16, 2023
 *      Author: algin
 */

#include "ethernet_industry/CoE/eth_ecat_sdo_mailbox.hpp"

namespace free_rtos {

namespace ecat_sdo_mailbox {

void EthEcatSdoMailbox::init() {
    std::vector<ecat_buffer::EcatBufferSlave>& buffer_slaves = ecat_buffer_.get_buffer_slaves();

    sdo_mailbox_slaves_.reserve(buffer_slaves.size());

    for(ecat_buffer::EcatBufferSlave& mailbox_slave : buffer_slaves) {
        sdo_mailbox_slaves_.push_back(EcatSdoMailboxSlave{mailbox_slave});
    }
}

void EthEcatSdoMailbox::pdo_map_read(PDOMap& pdo_map) {
    datagram::EcatTelegram& telegram = ecat_buffer_.get_ecat().get_telegram();
    uint16_t pdo_map_rx_index{0x1C12};
    uint16_t pdo_map_tx_index{0x1C13};
    uint16_t pdo_rx_data_size{0x0000};
    uint16_t pdo_tx_data_size{0x0000};

    for(EcatSdoMailboxSlave& sdo_mailbox_slave : sdo_mailbox_slaves_) {
        DebugP_log("Reading rx pdo map\r\n");
        pdo_rx_data_size += sdo_mailbox_slave.pdo_map_read<command::FP>(telegram, pdo_map_rx_index);

        DebugP_log("Reading tx pdo map\r\n");
        pdo_tx_data_size += sdo_mailbox_slave.pdo_map_read<command::FP>(telegram, pdo_map_tx_index);
    }

    pdo_map.rx_data_size = pdo_rx_data_size;
    pdo_map.tx_data_size = pdo_tx_data_size;

    DebugP_log("pdo_map.rx_data_size = %d\r\n", pdo_map.rx_data_size);
    DebugP_log("pdo_map.tx_data_size = %d\r\n", pdo_map.tx_data_size);
}

}

}
