/*
 * eth_ecat_custom_tuple.hpp
 *
 *  Created on: May 5, 2023
 *      Author: algin
 */

#ifndef FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_CUSTOM_TUPLE_HPP_
#define FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_CUSTOM_TUPLE_HPP_

namespace free_rtos {

// Базовый шаблон класса, никогда не инстанциируется, поэтому без тела
template<typename... Args>
struct CustomTuple;

// Основная специализация шаблона. Есть еще одна для протокола CoE.
template<typename HeadT, typename... TailT>
struct CustomTuple<HeadT, TailT...> : CustomTuple<TailT...> {
    CustomTuple(HeadT& head, TailT&... tail)
    : CustomTuple<TailT...>(tail...)
    , head_(head) { }

    using TBase = CustomTuple<TailT...>;

    constexpr static size_t size = sizeof(HeadT) + TBase::size;

    TBase& base_ = static_cast<TBase&>(*this);
    HeadT& head_;

    uint8_t* pack(uint8_t *raw) {
        HeadT *head = new(raw) HeadT{head_};

        (void)head;

        return base_.pack(raw + sizeof(HeadT));
    }

    uint8_t* unpack(uint8_t *raw) {
        HeadT *head = reinterpret_cast<HeadT*>(raw);

        head_ = *head;

        return base_.unpack(raw + sizeof(HeadT));
    }
};

// Специализация завершения рекурсии
template<>
struct CustomTuple<> {
    constexpr static size_t size = 0;

    uint8_t* pack(uint8_t *raw) {
        return raw;
    }

    uint8_t* unpack(uint8_t *raw) {
        return raw;
    }
};

}



#endif /* FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_CUSTOM_TUPLE_HPP_ */
